unit CodeCoop;
// This file is part of Sarah Dean's Code Co-op Shell Extension.
//
//  "Sarah Dean's Code Co-op Shell Extension" is free software; you can
// redistribute it and/or modify it under the terms of the GNU General
// Public License as published by the Free Software Foundation; either
// version 2 of the License, or at your option) any later version.
//
//  "Sarah Dean's Code Co-op Shell Extension" is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with "Sarah Dean's Code Co-op Shell Extension"; if not, write to
// the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
// MA  02111-1307  USA



interface

uses
  Classes,
  ConsoleApp;

type
    // Note: fileStatusUnknown will be returned for files not under a
    //       Code Co-op project tree.
    //       fileStatusNotUnderSCC will be returned for files which *are* under
    //       a Code Co-op project tree, but which haven't been added to
    //       Code Co-op
    TCCFileStatus = (fileStatusCheckedIn, fileStatusCheckedOut, fileStatusNotUnderSCC, fileStatusUnknown);
    TCCFileType = (fileTypeHeader, fileTypeSource, fileTypeText, fileTypeBinary);

type
    TCodeCoop = class
    protected
      fActive: boolean;

      function  GetCodeCoopCmdUtilDir(): string;

      procedure SetActive(state: boolean);

      function  CheckActive(): boolean;

    public
      constructor Create();
      destructor Destroy(); override;

      property Active: boolean read fActive write SetActive;

      function GetProjects(name: TStringList; rootDir: TStringList): boolean;

      function FileStatus(filename: string): TCCFileStatus;

      function AddFile(filename: string; fileType: TCCFileType): boolean;
      // If "filename" is already checked in this function will return success
      // If "filename" is not under Code Co-op, this function will fail
      function CheckIn(filename: string; keepCheckedOut: boolean; comment: string): boolean;
      // If "filename" is already checked out, this function will return true
      // If "filename" is not under Code Co-op, this function will fail
      function CheckOut(filename: string): boolean;
      // If "filename" is already checked in, this function will return true
      // If "filename" is not under Code Co-op, this function will fail
      function UncheckOut(filename: string): boolean;


    end;

implementation

uses
  Windows, SysUtils, Registry,
  SDUGeneral;


const
  CODE_COOP_PROJECTSTATUS = 'ProjectStatus.exe';

  CODE_COOP_FILESTATUS    = 'Status.exe';

  CODE_COOP_ADDFILE       = 'addfile.exe';
  CODE_COOP_CHECKIN       = 'checkin.exe';
  CODE_COOP_CHECKOUT      = 'checkout.exe';
  CODE_COOP_UNCHECKOUT    = 'uncheckout.exe';


  // Registry key storing the directory in which the Code Co-op command line
  // utilities have been installed
  REGKEY_CODECOOPCMDLINEUTILS_ROOT = HKEY_LOCAL_MACHINE;
  REGKEY_CODECOOPCMDLINEUTILS_KEY = '\SOFTWARE\Reliable Software\Code Co-op';
  REGKEY_CODECOOPCMDLINEUTILS_NAME = 'CmdLineToolsPath';



constructor TCodeCoop.Create();
begin
  inherited;

  fActive := FALSE;
end;


destructor TCodeCoop.Destroy();
begin
  inherited;
end;


procedure TCodeCoop.SetActive(state: boolean);
var
  projectNames: TStringList;
  projectRootDirs: TStringList;
begin
  fActive := FALSE;
  if (state) then
    begin
    // Check that we can get the command line utilities dir
    if (GetCodeCoopCmdUtilDir()<>'') then
      begin
      // Check that we can get a list of projects
      projectNames:= TStringList.Create();
      try
        projectRootDirs:= TStringList.Create();
        try
          // Set fActive to TRUE, just so that GetProjects(...) can be tested 
          fActive := TRUE;
          fActive := GetProjects(projectNames, projectRootDirs);

        finally
          projectRootDirs.Free();
        end;

      finally
        projectNames.Free();
      end;

      end;
    
    end
  else
    begin
    fActive := state;
    end;

end;


function TCodeCoop.CheckActive(): boolean;
begin
  if not(fActive) then
    begin
    raise Exception.Create('Code Co-op component not connected - set Active to TRUE first');
    end;

  Result := fActive;

end;




function TCodeCoop.GetProjects(name: TStringList; rootDir: TStringList): boolean;
var
  exeFilename: string;
  stdout: TStringList;
  i: integer;
  retval: boolean;
  projectName: string;
  projectDir: string;
  eval: DWORD;
  tmpProjectDir: string;
  junk: string;
begin
  retval := FALSE;

  CheckActive();

  exeFilename:= GetCodeCoopCmdUtilDir()+'\'+CODE_COOP_PROJECTSTATUS;

  stdout:= TStringList.Create();
  try
    try
      eval := ExecConsoleApp(exeFilename,
                             '-a',
                             stdout,
                             nil
                            );
    except
      eval := 999;
    end;

    if (eval=0) then
      begin
      // All lines in the std should have the format:
      //   <project name> (<project dir> -- <n> checked out file(s);
      //   <project name> (<project dir>)
      if (stdout.count>1) then
        begin
        retval := TRUE;
        for i:=0 to (stdout.Count-1) do
          begin
          if stdout[i]<>'' then
            begin
            if (pos('(', stdout[0])>0) then
              begin
              // Get the project name
              if SDUSplitString(stdout[i], projectName, projectDir, '(') then
                begin
                // Trim the extra space between the project name, and what was
                // the "("
                delete(projectName, length(projectName), 1);

                if (name<>nil) then
                  begin
                  name.Add(projectName);
                  end;

                // Attempt to determine the project directory
                if (rootDir<>nil) then
                  begin
                  // Attempt to determine the project directory
                  tmpProjectDir := projectDir;
                  if SDUSplitString(tmpProjectDir, projectDir, junk, ')') then
                    begin
                    rootDir.Add(projectDir);
                    end
                  else
                    begin
                    retval := FALSE;
                    break;
                    end;
                  end;
                end
              else
                begin
                retval := FALSE;
                break;
                end;

              end
            else
              begin
              retval := FALSE;
              break;
              end;
            end;
          end;
        end;
      end;

  finally
    stdout.Free();
  end;


  Result := retval;
end;


function TCodeCoop.FileStatus(filename: string): TCCFileStatus;
var
  retval: TCCFileStatus;
  exeFilename: string;
  stdout: TStringList;
  eval: DWORD;
  ucFilename: string;
  filenameIsRootDir: boolean;
  projectRootDirs: TStringList;
  i: integer;
begin
  retval := fileStatusUnknown;

  CheckActive();

  exeFilename:= GetCodeCoopCmdUtilDir()+'\'+CODE_COOP_FILESTATUS;

  stdout:= TStringList.Create();
  try
    try
      eval := ExecConsoleApp(exeFilename,
                             '"'+filename+'"',
                             stdout,
                             nil
                            );
    except
      eval := 999;
    end;

    if (eval=0) then
      begin
      // All lines in the std should have the format:
      //   <filename> - <status text>
      if (stdout.count=2) then
        begin
        if (pos(' - checked in', stdout[0])>0) then
          begin
          // Sanity check - the dir in which a Code Co-op project is located
          // gets reported as "Checked in" by the command line utils; even
          // though it shouldn't be.
          filenameIsRootDir := FALSE;
          projectRootDirs:= TStringList.Create();
          try
            if GetProjects(nil, projectRootDirs) then
              begin
              ucFilename:= uppercase(filename);
              for i:=0 to (projectRootDirs.count-1) do
                begin
                if (uppercase(projectRootDirs[i]) = ucFilename) then
                  begin
                  filenameIsRootDir := TRUE;
                  break;
                  end;
                end;
              end;

          finally
            projectRootDirs.Free();
          end;


          if not(filenameIsRootDir) then
            begin
            retval := fileStatusCheckedIn;
            end;
          end
        else if (pos(' - checked out', stdout[0])>0) then
          begin
          retval := fileStatusCheckedOut;
          end
        else if (pos(' - NOT a project file', stdout[0])>0) then
          begin
          // Determine if the file is under a Code Co-op project tree
          projectRootDirs:= TStringList.Create();
          try
            if GetProjects(nil, projectRootDirs) then
              begin
              ucFilename:= uppercase(filename);
              for i:=0 to (projectRootDirs.count-1) do
                begin
                if (Pos(uppercase(projectRootDirs[i]), ucFilename)>0) then
                  begin
                  retval := fileStatusNotUnderSCC;
                  break;
                  end;
                end;
              end;

          finally
            projectRootDirs.Free();
          end;

          end

        end;
      end;

  finally
    stdout.Free();
  end;


  Result := retval;
end;


// If "filename" is already under Code Co-op, this function will fail
function TCodeCoop.AddFile(filename: string; fileType: TCCFileType): boolean;
var
  retval: boolean;
  status: TCCFileStatus;
  exeFilename: string;
  stdout: TStringList;
  eval: DWORD;
  strFileType: string;
begin
  retval := FALSE;

  CheckActive();

  status:= FileStatus(filename);
  if (status = fileStatusCheckedIn) then
    begin
    retval := TRUE;
    end
  else if (status = fileStatusNotUnderSCC) then
    begin
    case fileType of
      fileTypeHeader:
        begin
        strFileType:= 'header';
        end;

      fileTypeSource:
        begin
        strFileType:= 'source';
        end;

      fileTypeText:
        begin
        strFileType:= 'text';
        end;

      fileTypeBinary:
        begin
        strFileType:= 'binary';
        end;

    end;

    exeFilename:= GetCodeCoopCmdUtilDir()+'\'+CODE_COOP_ADDFILE;


    stdout:= TStringList.Create();
    try
      try
        eval := ExecConsoleApp(exeFilename,
                               '-t:'+strFileType+' "'+filename+'"',
                               stdout,
                               nil
                              );
      except
        eval := 999;
      end;

      if (eval=0) then
        begin
        retval := TRUE;
        end;

    finally
      stdout.Free();
    end;
    
  end;


  Result:= retval;
end;


// If "filename" is already checked in this function will return success
// If "filename" is not under Code Co-op, this function will fail
function TCodeCoop.CheckIn(filename: string; keepCheckedOut: boolean; comment: string): boolean;
var
  retval: boolean;
  status: TCCFileStatus;
  exeFilename: string;
  stdout: TStringList;
  eval: DWORD;
  strOptKeepCheckedOut: string;
begin
  retval := FALSE;

  CheckActive();

  status:= FileStatus(filename);
  if (status = fileStatusCheckedIn) then
    begin
    retval := TRUE;
    end
  else if (status = fileStatusCheckedOut) then
    begin
    exeFilename:= GetCodeCoopCmdUtilDir()+'\'+CODE_COOP_CHECKIN;

    if keepCheckedOut then
      begin
      strOptKeepCheckedOut := '-o';
      end;

    stdout:= TStringList.Create();
    try
      try
        eval := ExecConsoleApp(exeFilename,
                               strOptKeepCheckedOut+' -c:"'+comment+'" "'+filename+'"',
                               stdout,
                               nil
                              );
      except
        eval := 999;
      end;

      if (eval=0) then
        begin
        retval := TRUE;
        end;

    finally
      stdout.Free();
    end;
    
  end;


  Result:= retval;
end;


// If "filename" is already checked out, this function will return true
// If "filename" is not under Code Co-op, this function will fail
function TCodeCoop.CheckOut(filename: string): boolean;
var
  retval: boolean;
  status: TCCFileStatus;
  exeFilename: string;
  stdout: TStringList;
  eval: DWORD;
begin
  retval := FALSE;

  CheckActive();

  status:= FileStatus(filename);
  if (status = fileStatusCheckedOut) then
    begin
    retval := TRUE;
    end
  else if (status = fileStatusCheckedIn) then
    begin
    exeFilename:= GetCodeCoopCmdUtilDir()+'\'+CODE_COOP_CHECKOUT;

    stdout:= TStringList.Create();
    try
      try
        eval := ExecConsoleApp(exeFilename,
                               '"'+filename+'"',
                               stdout,
                               nil
                              );
      except
        eval := 999;
      end;

      if (eval=0) then
        begin
        retval := TRUE;
        end;

    finally
      stdout.Free();
    end;
    
  end;


  Result:= retval;
end;


function TCodeCoop.UncheckOut(filename: string): boolean;
var
  retval: boolean;
  status: TCCFileStatus;
  exeFilename: string;
  stdout: TStringList;
  eval: DWORD;
begin
  retval := FALSE;

  CheckActive();

  status:= FileStatus(filename);
  if (status = fileStatusCheckedIn) then
    begin
    retval := TRUE;
    end
  else if (status = fileStatusCheckedOut) then
    begin
    exeFilename:= GetCodeCoopCmdUtilDir()+'\'+CODE_COOP_UNCHECKOUT;

    stdout:= TStringList.Create();
    try
      try
        eval := ExecConsoleApp(exeFilename,
                               '"'+filename+'"',
                               stdout,
                               nil
                              );
      except
        eval := 999;
      end;

      if (eval=0) then
        begin
        retval := TRUE;
        end;

    finally
      stdout.Free();
    end;

  end;


  Result:= retval;
end;



// Get the directory where the Code Co-op command line utilities are installed
// Returns '' on failure
function TCodeCoop.GetCodeCoopCmdUtilDir(): string;
var
  registry: TRegistry;
  retval: string;
begin
  retval := '';

  registry := TRegistry.create();
  try
    registry.RootKey := REGKEY_CODECOOPCMDLINEUTILS_ROOT;
    try
      begin
      // All extensions
      registry.OpenKey(REGKEY_CODECOOPCMDLINEUTILS_KEY, FALSE);
      retval := registry.ReadString(REGKEY_CODECOOPCMDLINEUTILS_NAME);
      registry.CloseKey();
      end;

    except
      on ERegistryException do
        begin
        retval := '';
        end;
    end;

  finally
    registry.Free();
  end;

  Result := retval;
end;


END.


